from functools import partial

from fastapi import FastAPI, Request, Response

from dropland.blocks import async_resource_session
from dropland.data.context import with_context


def add_middleware(
        app: FastAPI, begin_sql_tx: bool = True, autocommit: bool = True):
    @app.middleware('http')
    async def resource_session_middleware(request: Request, call_next) -> Response:
        with with_context():
            return await async_resource_session(
                partial(call_next, request), begin_tx=begin_sql_tx, autocommit=autocommit
            )



import contextlib
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Tuple

from contextvars import ContextVar

from dropland.blocks.sql.base import SqlStorageType
from .engine import EngineKey, SqlStorageEngine, SqlStorageBackend


@dataclass
class Session:
    engine: SqlStorageEngine
    connection: Any
    timeout_secs: int
    in_transaction: bool = False


class ConnectionContext:
    def __init__(self):
        self.sessions: Dict[str, Session] = dict()
        self.sessions_by_key: Dict[EngineKey, Set[str]] = defaultdict(set)


class SessionManager:
    def __init__(self, engine_factory: SqlStorageBackend):
        self._ctx: ContextVar[ConnectionContext] = ContextVar('_ctx', default=ConnectionContext())
        self._engine_factory = engine_factory

    def get_session(self, name: str) -> Optional[Session]:
        return self._ctx.get().sessions.get(name)

    def get_sessions_for_type(self, db_type: SqlStorageType, is_async: bool) -> List[Session]:
        ctx = self._ctx.get()
        conn_names = ctx.sessions_by_key[EngineKey(db_type=db_type, is_async=is_async)]
        return [ctx.sessions.get(name) for name in conn_names if name in ctx.sessions]

    def get_or_create_session(self, name: str) -> Tuple[bool, Optional[Session]]:
        if conn := self.get_session(name):
            return False, conn

        if engine := self._engine_factory.get_engine(name):
            return True, Session(
                engine=engine, connection=engine.new_connection(),
                timeout_secs=int(engine.timeout.total_seconds()))

        return False, None

    @contextlib.contextmanager
    def session_context(self, name: str, begin_tx: bool = True, autocommit: bool = True):
        created, session = self.get_or_create_session(name)
        assert not session.engine.is_async, \
            f'Sql engine with name "{name}" has only async driver, ' \
            f'use async_session_context() function instead'

        if not created:
            if begin_tx and not session.in_transaction:
                with session.engine.transaction_context(session.connection, autocommit):
                    try:
                        yield self._add_session(name, replace(session, in_transaction=True))
                    finally:
                        self._add_session(name, replace(session, in_transaction=False))
            else:
                yield session
            return

        with session.connection as conn:
            assert isinstance(conn, session.engine.connection_class), \
                f'Engine with name "{name}" has only async driver, ' \
                f'use async_session_context() function instead'

            session = replace(session, connection=conn, in_transaction=begin_tx)
            try:
                if begin_tx:
                    with session.engine.transaction_context(session.connection, autocommit):
                        yield self._add_session(name, session)
                else:
                    yield self._add_session(name, session)
            finally:
                self._remove_session(name)

    @contextlib.asynccontextmanager
    async def async_session_context(self, name: str, begin_tx: bool = True, autocommit: bool = True):
        created, session = self.get_or_create_session(name)
        assert session.engine.is_async, \
            f'Sql engine with name "{name}" has only sync driver, ' \
            f'use session_context() function instead'

        if not created:
            if begin_tx and not session.in_transaction:
                async with session.engine.async_transaction_context(session.connection, autocommit):
                    try:
                        yield self._add_session(name, replace(session, in_transaction=True))
                    finally:
                        self._add_session(name, replace(session, in_transaction=False))
            else:
                yield session
            return

        async with session.connection as conn:
            assert isinstance(conn, session.engine.connection_class), \
                f'Sql engine with name "{name}" has only sync driver, ' \
                f'use session_context() function instead'

            session = replace(session, connection=conn, in_transaction=begin_tx)
            try:
                if begin_tx:
                    async with session.engine.async_transaction_context(session.connection, autocommit):
                        yield self._add_session(name, session)
                else:
                    yield self._add_session(name, session)
            finally:
                self._remove_session(name)

    @contextlib.contextmanager
    def init_engines(self, engines: Mapping[str, Sequence[SqlStorageEngine]] = None,
                     begin_tx: bool = True, autocommit: bool = True):
        engines = self._engine_factory.get_engines(engines or [])
        with contextlib.ExitStack() as stack:
            for name, engine in engines.items():
                if engine.is_async:
                    continue
                engine.start()
                stack.callback(engine.stop)
                stack.enter_context(self.session_context(name, begin_tx, autocommit))

            yield self._ctx.get()

    @contextlib.asynccontextmanager
    async def init_async_engines(self, engines: Mapping[str, Sequence[SqlStorageEngine]] = None,
                                 begin_tx: bool = True, autocommit: bool = True):
        engines = self._engine_factory.get_engines(engines or [])
        async with contextlib.AsyncExitStack() as stack:
            for name, engine in engines.items():
                if engine.is_async:
                    await engine.async_start()
                    stack.push_async_callback(engine.async_stop)
                    await stack.enter_async_context(
                        self.async_session_context(name, begin_tx, autocommit)
                    )
                else:
                    engine.start()
                    stack.callback(engine.stop)
                    stack.enter_context(self.session_context(name, begin_tx, autocommit))

            yield self._ctx.get()

    def _add_session(self, name: str, data: Session) -> Session:
        ctx = self._ctx.get()
        ctx.sessions[name] = data
        ctx.sessions_by_key[EngineKey(db_type=data.engine.db_type, is_async=data.engine.is_async)].add(name)
        return data

    def _remove_session(self, name: str):
        ctx = self._ctx.get()
        data = ctx.sessions.pop(name, None)
        ctx.sessions_by_key.pop(EngineKey(db_type=data.engine.db_type, is_async=data.engine.is_async), None)



from functools import partial

from dropland.util import invoke_async, invoke_sync

from .redis import USE_REDIS
from .rmq import USE_RMQ
from .sql import USE_SQL


def sync_resource_session(func, *args, **kwargs):
    if USE_SQL:
        from .sql.containers import sync_resource_session

        func = partial(sync_resource_session, func, *args, **kwargs)

    return invoke_sync(func)


async def async_resource_session(func, *args, **kwargs):
    if USE_RMQ:
        from .rmq.containers import async_resource_session

        func = partial(async_resource_session, func, *args, **kwargs)

    if USE_REDIS:
        from .redis.containers import async_resource_session

        func = partial(async_resource_session, func, *args, **kwargs)

    if USE_SQL:
        from .sql.containers import async_resource_session

        func = partial(async_resource_session, func, *args, **kwargs)

    return await invoke_async(func)



from dependency_injector import containers, providers
from dependency_injector.wiring import Provide, inject

from dropland.util import default_value, invoke_async, invoke_sync
from .engine import EngineConfig, SqlStorageBackend
from .session import SessionManager


class SqlStorage(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(SqlStorageBackend)
    manager = providers.Singleton(SessionManager, engine_factory)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    def _init_session(self, begin_tx: bool = True, autocommit: bool = True):
        with self.manager().init_engines(begin_tx=begin_tx, autocommit=autocommit):
            yield self.manager()

    async def _init_async_session(self, begin_tx: bool = True, autocommit: bool = True):
        async with self.manager().init_async_engines(begin_tx=begin_tx, autocommit=autocommit):
            yield self.manager()

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(_init_session, __self__)
    async_session_context = providers.Resource(_init_async_session, __self__)
    get_session = providers.Factory(manager.provided.get_session.call())

    wiring_config = containers.WiringConfiguration(
        modules=['.model', __name__]
    )


class SingleSqlStorage(SqlStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                echo=self.config.engine_config.echo.as_(bool)(),
                pool_min_size=self.config.engine_config.
                    pool_min_size.as_(default_value(int))(default=1),
                pool_max_size=self.config.engine_config.
                    pool_max_size.as_(default_value(int))(default=8),
                pool_expire_seconds=self.config.engine_config.
                    pool_expire_seconds.as_(default_value(int))(default=60),
                pool_timeout_seconds=self.config.engine_config.
                    pool_timeout_seconds.as_(default_value(int))(default=15)
            )
        return SqlStorage._create_engine(
            self, self.config.name(), engine_config, self.config.db_type(),
            use_async=self.config.use_async.as_(bool)())

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(SqlStorage._init_session, __self__)
    async_session_context = providers.Resource(SqlStorage._init_async_session, __self__)
    get_session = providers.Factory(SqlStorage.get_session, config.name)

    wiring_config = containers.WiringConfiguration(
        modules=['.model', __name__]
    )


class MultipleSqlStorage(SqlStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    echo=bool(conf['engine_config'].get('echo')),
                    pool_min_size=int(conf['engine_config'].get('pool_min_size', 1)),
                    pool_max_size=int(conf['engine_config'].get('pool_max_size', 8)),
                    pool_expire_seconds=int(conf['engine_config'].get('pool_expire_seconds', 60)),
                    pool_timeout_seconds=int(conf['engine_config'].get('pool_timeout_seconds', 15))
                )
            return SqlStorage._create_engine(
                self, name, engine_config, conf['db_type'], use_async=conf.get('use_async', False))
        return None

    def _get_session(self, name: str):
        return self.manager().get_session(name)

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(SqlStorage._init_session, __self__)
    async_session_context = providers.Resource(SqlStorage._init_async_session, __self__)
    get_session = providers.Factory(_get_session, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=['.model', __name__]
    )


default_sql_storage = SqlStorage()


@inject
def sync_resource_session(func, begin_tx: bool = True, autocommit: bool = True,
                          resource: SqlStorage = Provide['<container>'], *args, **kwargs):
    if not isinstance(resource, (SqlStorage, containers.DynamicContainer)):
        return invoke_sync(func)

    try:
        resource.session_context(begin_tx=begin_tx, autocommit=autocommit)
        return invoke_sync(func)
    finally:
        resource.session_context.shutdown()


@inject
async def async_resource_session(func, begin_tx: bool = True, autocommit: bool = True,
                                 resource: SqlStorage = Provide['<container>'], *args, **kwargs):
    if not isinstance(resource, (SqlStorage, containers.DynamicContainer)):
        return await invoke_async(func)

    try:
        await resource.async_session_context(begin_tx=begin_tx, autocommit=autocommit)
        return await invoke_async(func)
    finally:
        await resource.async_session_context.shutdown()



from dependency_injector import containers, providers
from dependency_injector.wiring import Provide, inject
from dropland.util import default_value, invoke_async

from .engine import EngineConfig, RmqStorageBackend
from .session import SessionManager


class RmqStorage(containers.DeclarativeContainer):
    __self__ = providers.Self()
    engine_factory = providers.Singleton(RmqStorageBackend)
    manager = providers.Singleton(SessionManager, engine_factory)

    def _create_engine(self, *args, **kwargs):
        return self.engine_factory().create_engine(*args, **kwargs)

    async def _init_async_session(self):
        async with self.manager().init_async_engines():
            yield self.manager()

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(_init_async_session, __self__)
    get_session = providers.Factory(manager.provided.get_session.call())

    wiring_config = containers.WiringConfiguration(
        modules=[__name__]
    )


class SingleRmqStorage(RmqStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self):
        if isinstance(self.config.engine_config(), EngineConfig):
            engine_config = self.config.engine_config()
        else:
            engine_config = EngineConfig(
                url=self.config.engine_config.url(),
                virtualhost=self.config.engine_config.virtualhost(),
                timeout_seconds=self.config.engine_config.
                    timeout_seconds.as_(default_value(int))(default=5),
                pool_max_connections=self.config.engine_config.
                    pool_max_connections.as_(default_value(int))(default=4),
                pool_max_channels_per_connection=self.config.engine_config.
                    pool_max_channels_per_connection.as_(default_value(int))(default=100)
            )
        return RmqStorage._create_engine(self, self.config.name(), engine_config)

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(RmqStorage._init_async_session, __self__)
    get_session = providers.Factory(RmqStorage.get_session, config.name)

    wiring_config = containers.WiringConfiguration(
        modules=[__name__]
    )


class MultipleRmqStorage(RmqStorage):
    __self__ = providers.Self()
    config = providers.Configuration()

    def _create_engine(self, name: str):
        if conf := self.config.get(name):
            if isinstance(conf['engine_config'], EngineConfig):
                engine_config = conf['engine_config']
            else:
                engine_config = EngineConfig(
                    url=conf['engine_config']['url'],
                    virtualhost=conf['engine_config']['virtualhost'],
                    timeout_seconds=int(conf['engine_config'].get('timeout_seconds', 5)),
                    pool_max_connections=int(conf['engine_config'].get('pool_max_connections', 4)),
                    pool_max_channels_per_connection=int(conf['engine_config'].get(
                        'pool_max_channels_per_connection', 100))
                )
            return RmqStorage._create_engine(self, name, engine_config)
        return None

    def _get_session(self, name: str):
        return self.manager().get_session(name)

    create_engine = providers.Factory(_create_engine, __self__)
    session_context = providers.Resource(RmqStorage._init_async_session, __self__)
    get_session = providers.Factory(_get_session, __self__)

    wiring_config = containers.WiringConfiguration(
        modules=[__name__]
    )


default_rmq_storage = RmqStorage()


@inject
async def async_resource_session(func, resource: RmqStorage = Provide['<container>'], *args, **kwargs):
    if not isinstance(resource, (RmqStorage, containers.DynamicContainer)):
        return await invoke_async(func)

    try:
        await resource.session_context.init()
        return await invoke_async(func)
    finally:
        await resource.session_context.shutdown()



import contextlib
from dataclasses import dataclass, replace
from typing import Dict, List, Optional, Tuple

from aio_pika import RobustConnection as RmqConnection
from contextvars import ContextVar

from .engine import RmqStorageBackend, RmqStorageEngine


@dataclass
class Session:
    engine: RmqStorageEngine
    connection: RmqConnection


class ConnectionContext:
    def __init__(self):
        self.sessions: Dict[str, Session] = dict()


class SessionManager:
    def __init__(self, engine_factory: RmqStorageBackend):
        self._ctx: ContextVar[ConnectionContext] = ContextVar('_ctx', default=ConnectionContext())
        self._engine_factory = engine_factory

    def get_session(self, name: str) -> Optional[Session]:
        return self._ctx.get().sessions.get(name)

    def get_or_create_session(self, name: str) -> Tuple[bool, Optional[Session]]:
        if conn := self.get_session(name):
            return False, conn

        if engine := self._engine_factory.get_engine(name):
            return True, Session(
                engine=engine,
                connection=engine.new_connection())

        return False, None

    @contextlib.asynccontextmanager
    async def async_session_context(self, name: str):
        created, data = self.get_or_create_session(name)

        if not created:
            yield data
            return

        async with data.connection as conn:
            data = replace(data, connection=conn)
            try:
                yield self._add_session(name, data)
            finally:
                self._remove_session(name)

    @contextlib.asynccontextmanager
    async def init_async_engines(self, names: List[str] = None):
        engines = self._engine_factory.get_engines(names or [])
        async with contextlib.AsyncExitStack() as stack:
            for name, engine in engines.items():
                assert engine is self._engine_factory.get_engine(name)
                if engine.is_async:
                    await engine.async_start()
                    stack.push_async_callback(engine.async_stop)
                else:
                    engine.start()
                    stack.callback(engine.stop)

                await stack.enter_async_context(self.async_session_context(name))

            yield self._ctx.get()

    def _add_session(self, name: str, data: Session) -> Session:
        self._ctx.get().sessions[name] = data
        return data

    def _remove_session(self, name: str):
        self._ctx.get().sessions.pop(name)
